from .base_driver import BaseOCRDriver, WindowContext
from .linux_driver import LinuxOCRDriver


def get_driver() -> BaseOCRDriver:
    """In the future, we can return other OS compatible OCR solutions here"""
    return LinuxOCRDriver()



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional

_MIN_CHARACTERS_PER_LINE = 10


@dataclass
class WindowContext:
    raw_screen_text: Optional[str] = None
    """If the driver supports it, the text extracted from the active window will be
    filled here."""

    active_window_name: Optional[str] = None
    """If the driver supports it, the current active window name will be filled here."""

    @property
    def clean_screen_text(self) -> Optional[str]:
        if not self.raw_screen_text:
            return None

        lines: list[str] = self.raw_screen_text.split("\n")
        clean_lines = []
        for line in lines:
            line = line.strip()
            if len(line) > _MIN_CHARACTERS_PER_LINE:
                clean_lines.append(line)

        clean_text = "\n".join(clean_lines)
        return clean_text


class BaseOCRDriver(ABC):
    """This base class can be used to standardize the interface for future OS's"""

    @abstractmethod
    def extract_context(self) -> WindowContext:
        """A method to extract the current useful context from the in-focus window"""
        pass



from typing import cast

import pyautogui
import pytesseract
import pywinctl
from PIL.Image import Image

from .base_driver import BaseOCRDriver, WindowContext


class LinuxOCRDriver(BaseOCRDriver):
    def extract_context(self) -> WindowContext:
        screenshot = self._extract_active_window_screenshot()

        # Perform OCR and clean up the text
        raw_ocr_text = pytesseract.image_to_string(screenshot)

        return WindowContext(
            raw_screen_text=raw_ocr_text,
            active_window_name=cast(str, pywinctl.getActiveWindowTitle()),
        )

    def _extract_active_window_screenshot(self) -> Image:
        # Get the active window object
        active_window = pywinctl.getActiveWindow()  # type: ignore
        region = (
            active_window.left,
            active_window.top,
            active_window.width,
            active_window.height,
        )

        # Take a screenshot of the active window
        screenshot = pyautogui.screenshot(region=region)
        return screenshot



import os
from types import ModuleType

import openai

API_TOKEN_VAR = "OPENAI_API_TOKEN"


def initialize_api() -> ModuleType:
    try:
        api_key = os.environ[API_TOKEN_VAR]
    except KeyError:
        print(f"You must set the`{API_TOKEN_VAR}` variable in your environment!")
        exit(1)

    openai.api_key = api_key
    return openai



from .behavior_context import MESSAGE_CONTEXT, Message, Prompt
from .ocr_drivers import get_driver


def create_message_context(prompt: str) -> list[dict[str, str]]:
    """An important part of a ChatBot is the prior context. Here, we carefully
    construct the context for the messages we are sending.

    This function returns an OpenAI-API ready message with the prompt included.
    :param prompt: The prompt for the chatbot
    :return: The API-ready message
    """
    driver = get_driver()
    window_context = driver.extract_context()
    new_context = MESSAGE_CONTEXT[:]
    new_context.append(
        Message(role="user", content=Prompt(context=window_context, prompt=prompt))
    )
    api_format = [m.to_api() for m in new_context]
    return api_format



from dataclasses import dataclass
from typing import Literal, Union

from clai.ocr_drivers import WindowContext

USER_PROMPT_FORMAT = """
User Prompt:
```
{user_prompt}
```
"""
OCR_EXTRACTION_FORMAT = """
Active Window Title: {active_window_title}

Active Window OCR Extracted Text (RAW):
------ OCR DATA START ------
```
{ocr_text}
```
------ OCR DATA END ------

{user_prompt}

Please answer "User Prompt" using the raw OCR text as context to the message.
"""


@dataclass
class Prompt:
    context: WindowContext
    prompt: str

    def __str__(self) -> str:
        """Serialize the Prompt with differing formats, depending on whether window
        content was available

        :return: The window context and prompt in a standardized format
        """
        """."""
        user_prompt = USER_PROMPT_FORMAT.format(user_prompt=self.prompt.strip())
        if self.context.clean_screen_text and self.context.active_window_name:
            return OCR_EXTRACTION_FORMAT.format(
                active_window_title=self.context.active_window_name.strip(),
                ocr_text=self.context.clean_screen_text.strip(),
                user_prompt=user_prompt.strip(),
            )
        return user_prompt.strip()


@dataclass
class Message:
    role: Literal["system", "user", "assistant"]
    content: Union[Prompt, str]

    def to_api(self) -> dict[str, str]:
        """To OpenAPI format"""
        if isinstance(self.content, str) and self.role == "user":
            raise RuntimeError("The user message must be of type Prompt!")

        return {"role": self.role, "content": str(self.content)}


_DEFAULT_ASSISTANT_ROLE = """
You are an assistant that is capable of being called anywhere on a desktop computer. You
may be called within the context of an email, a URL box, commandline, a text editor, or
even word documents!

Your role is to answer the users request as shortly and succinctly as possible. You
will follow the following rules:

When asked to write long-form text content:
1) Never ask for more information. If something is to be guessed, write it in template
   format. For example, if asked to write an email use <Insert time here> when writing
   the portion of an email that specifies something that was not included in the users
   question.
2) Only assume the content is long form if the user mentions email, or 'long message'.

When asked to write a command, code, formulas, or any one-line response task:
1) NEVER WRITE EXPLANATIONS! Only include the command/code/etc, ready to be run
2) NEVER WRITE USAGE INSTRUCTIONS! Do not explain how to use the command/code/formulas.
3) NEVER WRITE NOTES ABOUT THE IMPLEMENTATION!
   Do not explain what it does or it's limitations.
4) Remember, the text that you write will immediately be run, do not include code blocks
5) If there is something that requires user input, such as a cell in a sheet or a
   variable from the user, write it inside of brackets, like this: <INPUT DESCRIBER>,
   where the insides of the bracket have an example of what is needed to be filled in.
6) Assume a linux desktop environment in a bash shell. Use freely available unix tools.

You will receive OCR context and window title names, for some prompts. They are very 
noisy, use best-effort when reading them.
"""
_EXAMPLE_EMAIL = """
Dear <Recipient's Name>,

I hope this email finds you well. I am writing to request a meeting with you on <Date and Time>, and I would appreciate it if you could confirm your availability at your earliest convenience.

The purpose of this meeting is to discuss <Purpose of the Meeting> with you. Specifically, I would like to <Agenda Item 1>, <Agenda Item 2>, and <Agenda Item 3>. The meeting will last approximately <Meeting Duration> and will take place at <Meeting Location>.

Please let me know if this date and time work for you. If not, please suggest an alternative time that is convenient for you. Additionally, if there are any documents or information you would like me to review before the meeting, please let me know, and I will make sure to review them.

I look forward to hearing from you soon.

Best regards,

<Your Name>
"""  # noqa: E501
_EXAMPLE_REGEX = '=IFERROR(REGEXEXTRACT(<INPUT CELL HERE>, "[A-z0-9._%+-]+@[A-z0-9.-]+\.[A-z]{2,4}");"")'  # noqa
_EXAMPLE_PYTHON = """
def fibonacci(n: int) -> Generator[int, None, None]:
    a, b = 0, 1
    for _ in range(n):
        yield a
        a, b = b, a + b
"""
_EXAMPLE_GOOGLE_SHEETS = '=IFERROR(REGEXEXTRACT(<INPUT CELL HERE>, "[A-z0-9._%+-]+@[A-z0-9.-]+\.[A-z]{2,4}");"")'  # noqa
_EXAMPLE_BASH_COMMAND = "grep -rnw . -e 'bruh'"

MESSAGE_CONTEXT: list[Message] = [
    Message(role="system", content=_DEFAULT_ASSISTANT_ROLE),
    Message(
        role="user",
        content=Prompt(
            WindowContext(),
            prompt="commandline search for files with the name 'bruh' in them",
        ),
    ),
    Message(role="assistant", content=_EXAMPLE_BASH_COMMAND),
    Message(
        role="user",
        content=Prompt(
            context=WindowContext(), prompt="email set up a meeting next week"
        ),
    ),
    Message(role="assistant", content=_EXAMPLE_EMAIL),
    Message(
        role="user",
        content=Prompt(
            context=WindowContext(),
            prompt="google sheets formula extracts an email from string of text",
        ),
    ),
    Message(role="assistant", content=_EXAMPLE_GOOGLE_SHEETS),
    Message(
        role="user",
        content=Prompt(
            context=WindowContext(),
            prompt="google sheets formula extracts an email from string of text",
        ),
    ),
    Message(role="assistant", content=_EXAMPLE_REGEX),
    Message(
        role="user",
        content=Prompt(
            context=WindowContext(),
            prompt="python fibonacci function in form of a generator",
        ),
    ),
    Message(role="assistant", content=_EXAMPLE_PYTHON),
]

__all__ = ["MESSAGE_CONTEXT", "Message", "Prompt"]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['clai', 'clai.ocr_drivers']

package_data = \
{'': ['*']}

install_requires = \
['PyAutoGUI>=0.9.53,<0.10.0',
 'PyWinCtl>=0.0.43,<0.0.44',
 'openai>=0.27.0,<0.28.0',
 'pytesseract>=0.3.10,<0.4.0']

entry_points = \
{'console_scripts': ['ai = clai:main', 'clai = clai:main']}

setup_kwargs = {
    'name': 'clai',
    'version': '0.3.0',
    'description': 'Command Line AI- this tool lets you call ChatGPT from a CLI',
    'long_description': '# clai\nCommand Line AI- this tool lets you call ChatGPT from a CLI. \n\nI\'m designing this to be used in conjunction with a fork of [shin][shin], which will allow you\nto call `clai` from any textbox in your computer. Finally, ChatGPT everywhere!\n\nThe long-term vision for this project is to add support for extracting context. For example, it would\nread the current text on a window and be able to add to it, or answer questions about it.\n\n_________________\n\n[![PyPI version](https://badge.fury.io/py/clai.svg)](http://badge.fury.io/py/clai)\n[![Test Status](https://github.com/apockill/clai/workflows/Test/badge.svg?branch=main)](https://github.com/apockill/clai/actions?query=workflow%3ATest)\n[![Lint Status](https://github.com/apockill/clai/workflows/Lint/badge.svg?branch=main)](https://github.com/apockill/clai/actions?query=workflow%3ALint)\n[![codecov](https://codecov.io/gh/apockill/clai/branch/main/graph/badge.svg)](https://codecov.io/gh/apockill/clai)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![Imports: isort](https://img.shields.io/badge/%20imports-isort-%231674b1?style=flat&labelColor=ef8336)](https://timothycrosley.github.io/isort/)\n_________________\n\n[Read Latest Documentation](https://apockill.github.io/clai/) - [Browse GitHub Code Repository](https://github.com/apockill/clai/)\n_________________\n\n## Installation\n\n1. The recommended installation method is to use `pipx`, via\n    ```bash\n    pipx install clai\n    ```\n   Optionally, install `tesseract` so that `clai` can read the screen context and send that along with requests:\n   ```bash\n   sudo apt install tesseract-ocr scrot\n   ```\n1. Then go to [OpenAI] and create an API Key. Once it\'s generated, add the following to \n   your `~/.profile`:\n   ```bash\n   export OPENAI_API_TOKEN=<paste here>\n   ```\n\n1. The best way to use this tool is in conjunction with the tool [shin][shin], which allows you\n   to run arbitrary bash commands in any textbox in a linux computer, using ibus. To use \n   that, install \'shin\' via the fork above, then configure\n   it in your `~/.profile` to call `clai` by default:\n   ```bash\n   export SHIN_DEFAULT_COMMAND="clai"\n   ```\n1. Log out then log back in for the changes to take effect!\n\n[OpenAI]: https://platform.openai.com/account/api-keys\n\n## Usage\nInvoke the assistant with the format `clai <your prompt>`. For example:\n```\nclai Write an email saying I\'ll be late to work because I\'m working on commandline AIs\n```\n\n\n## Development\n\n### Installing python dependencies\n```shell\npoetry install\n```\n\n### Running Tests\n```shell\npytest .\n```\n\n### Formatting Code\n```shell\nbash .github/format.sh\n```\n\n### Linting\n```shell\nbash .github/check_lint.sh\n```\n\n[shin]: https://github.com/apockill/shin\n',
    'author': 'apockill',
    'author_email': 'apocthiel@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from argparse import ArgumentParser

from .api import initialize_api
from .message_creation import create_message_context


def main() -> None:
    parser = ArgumentParser("CLAI- your own command line AI!")
    parser.add_argument("prompt", type=str, nargs="+")
    parser.add_argument("-m", "--model", default="gpt-3.5-turbo")
    args = parser.parse_args()

    openai = initialize_api()

    prompt = " ".join(args.prompt)
    response = openai.ChatCompletion.create(
        model=args.model,
        messages=create_message_context(prompt),
    )

    best_response = response["choices"][0]["message"]["content"]
    print(best_response)
